#include <Arduino.h>
#include "constants.h"
#include "debug.h"
#include "stepper.h"

/*

 A4988 Stepper Motor Driver.

*/

#if F_CPU != 16000000
  #error "Unsupported F_CPU"
#endif

bool setupStepper() {
  pinMode(STEPPER_ENABLE, OUTPUT);
  pinMode(STEPPER_DIR, OUTPUT);
  pinMode(STEPPER_STEP, OUTPUT);

  setStepperEnabled(false);

  // set microstepping
  pinMode(STEPPER_MS1, OUTPUT);
  pinMode(STEPPER_MS2, OUTPUT);
  pinMode(STEPPER_MS3, OUTPUT);

  #if STEPPER_MICROSTEPS == 2
  digitalWrite(STEPPER_MS1, HIGH);
  digitalWrite(STEPPER_MS2, LOW);
  digitalWrite(STEPPER_MS3, LOW);
  #else
  #error "Unsupported microstepping"
  #endif

  // sanity check
  if (!((int) portOutputRegister(digitalPinToPort(STEPPER_STEP)) == (int) &STEPPER_STEP_PORT
    && digitalPinToBitMask(STEPPER_STEP) == STEPPER_STEP_BIT
    )) {
      Serial.println("Stepper step pin is misconfigured");
      return false;
  }

  // set up the timer
  noInterrupts();
  // reset timer settings
  TCCR1A = 0;
  TCCR1B = 0;
  TCNT1 = 0;
  TIMSK1 &= ~(1 << OCIE1A);
  // 256 prescaler (16MHz / 256 = 62.5kHz)
  TCCR1B |= (1 << CS12);
  // CTC mode (Clear Timer on Compare Match)
  TCCR1B |= (1 << WGM12);
  // interrupt at 62.5kHz / (OCR1A + 1)
  OCR1A = 39;
  interrupts();

  return true;
}

bool isTimerEnabled() {
  return TIMSK1 & (1 << OCIE1A);
}

void setTimerEnabled(bool enabled) {
  if (enabled) {
    TIMSK1 |= (1 << OCIE1A);
  } else {
    TIMSK1 &= ~(1 << OCIE1A);
  }
}

uint16_t getTimerCadence() {
  return OCR1A;
}

void setTimerCadence(uint16_t cadence) {
  noInterrupts();
  // if TCNT1 > OCR1A, it won't be reset until it reaches uint16_t's max value
  TCNT1 = min(TCNT1, cadence - 1);
  // interrupt at 62.5kHz / (OCR1A + 1)
  OCR1A = cadence;
  interrupts();
}

void setStepperEnabled(bool enabled) {
  digitalWrite(STEPPER_ENABLE, !enabled);
}

bool getStepperDirection() {
  return !digitalRead(STEPPER_DIR);
}

void setStepperDirection(bool dir) {
  digitalWrite(STEPPER_DIR, !dir);
}

inline void step() {
  // NOTE: minimum pulse width 1us high, 1us low
  STEPPER_STEP_PORT |= STEPPER_STEP_BIT;
  delayMicroseconds(1);
  STEPPER_STEP_PORT &= ~STEPPER_STEP_BIT;
}

ISR(TIMER1_COMPA_vect) {
  step();
}

void quickAccelerate() {
  setStepperEnabled(true);
  uint16_t cadence = STEPPER_MIN_CADANCE;
  setTimerCadence(cadence);
  setTimerEnabled(true);
  while (cadence > STEPPER_MAX_CADANCE) {
    delay(STEPPER_CADENCE_ACCEL_MS);
    cadence--;
    setTimerCadence(cadence);
  }
}
